
#include "irc.h"
#include "struct.h"
#include "dcc.h"
#include "ircaux.h"
#include "ctcp.h"
#include "cdcc.h"
#include "input.h"
#include "status.h"
#include "lastlog.h"
#include "screen.h"
#include "vars.h" 
#include "misc.h"
#include "output.h"
#include "module.h"
#include "hook.h"
#include "hash2.h"
#include "napster.h"
#include "bsdglob.h"
#include "modval.h"

#include <sys/stat.h>
#include <sys/ioctl.h>

GetFile *getfile_struct = NULL;


char *napster_status(void)
{
int upload = 0;
int download = 0;
GetFile *tmp;
char buffer[NAP_BUFFER_SIZE+1];
char tmpbuff[80];
double perc = 0.0;
	if (!get_dllint_var("napster_window"))
		return m_strdup(empty_string);
	sprintf(buffer, "%s [Sh:%lu/%3.2f%s] ", 
		nap_current_channel ? nap_current_channel : empty_string, 
		statistics.shared_files, _GMKv(statistics.shared_filesize), 
		_GMKs(statistics.shared_filesize));
	for (tmp = getfile_struct; tmp; tmp = tmp->next, download++)
	{
                perc = (100.0 * (((double)(tmp->received + tmp->resume)) / (double)tmp->filesize));
		sprintf(tmpbuff, "%3.1f%%%%", perc);
		if (download)
			strcat(buffer, ",");
		else
			strcat(buffer, " [G:");
		strcat(buffer, tmpbuff);
	}
	if (download)
		strcat(buffer, "]");
	for (tmp = napster_sendqueue; tmp; tmp = tmp->next, upload++)
	{
                perc = (100.0 * (((double)(tmp->received + tmp->resume)) / (double)tmp->filesize));
		sprintf(tmpbuff, "%3.1f%%%%", perc);
		if (upload)
			strcat(buffer, ",");
		else
			strcat(buffer, " [S:");
		strcat(buffer, tmpbuff);
	}
	if (upload)
		strcat(buffer, "]");
	sprintf(tmpbuff, " [U:%d/D:%d]", upload, download);
	strcat(buffer, tmpbuff);
	return m_strdup(buffer);
}

void nap_del(IrcCommandDll *intp, char *command, char *args, char *subargs, char *helparg)
{
int count = 1;
GetFile *tmp, *last = NULL;
int num;
char *t;
	if (!args && *args)
		return;
	while ((t = next_arg(args, &args)))
	{	
		count = 1;
		if (!(num = my_atol(t)))
			continue;
		for (tmp = getfile_struct; tmp; tmp = tmp->next, count++)
		{
			if (count == num)
			{
				nap_say("%s", cparse("Removing $0 [$1-]", "%s %s", tmp->nick, base_name(tmp->filename)));
				if (tmp->socket > 0)
					close_socketread(tmp->socket);
				if (tmp->write > 0)
					close(tmp->write);
				if (last)
					last->next = tmp->next;
				else	
					getfile_struct = tmp->next;
				new_free(&tmp->nick);
				new_free(&tmp->ip);
				new_free(&tmp->checksum);
				new_free(&tmp->filename);
				new_free(&tmp->realfile);
				new_free(&tmp);
				build_napster_status(NULL);
				return;
			}
			last = tmp;
		}
		last = NULL;
		for (tmp = napster_sendqueue; tmp; tmp = tmp->next, count++)
		{
			if (count == num)
			{
				nap_say("%s", cparse("Removing $0 [$1-]", "%s %s", tmp->nick, base_name(tmp->filename)));
				if (tmp->socket > 0)
					close_socketread(tmp->socket);
				if (tmp->write > 0)
					close(tmp->write);
				if (last)
					last->next = tmp->next;
				else	
					napster_sendqueue = tmp->next;
				new_free(&tmp->nick);
				new_free(&tmp->ip);
				new_free(&tmp->checksum);
				new_free(&tmp->filename);
				new_free(&tmp->realfile);
				new_free(&tmp);
				build_napster_status(NULL);
				return;
			}
			last = tmp;
		}
	}
}

void nap_glist (IrcCommandDll *intp, char *command, char *args, char *subargs, char *helparg)
{
int count = 1;
GetFile *sg = getfile_struct;
time_t snow = now;
char	*dformat = "%W#$[3]0%n %Y$4%n $[14]1 $[-6]2$3 $5/$6 $7-";

	for(sg = getfile_struct; sg; sg = sg->next, count++)
	{
		char buff[80];
		char buff1[80];
		char buff2[80];
		double perc = 0.0;
		if (count == 1)
		{
			nap_put("%s", cparse("%GDownloads", NULL));
			nap_put("%s", cparse("%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K", NULL, NULL));
		}
		if (sg->starttime)
			sprintf(buff, "%2.3f", sg->received / 1024.0 / (snow - sg->starttime));
		else
			strcpy(buff, "N/A");
		perc = (100.0 * (((double)(sg->received + sg->resume)) / (double)sg->filesize));
		sprintf(buff1, "%4.1f%%", perc);
		sprintf(buff2, "%4.2f", _GMKv(sg->filesize));
		nap_put("%s", cparse(dformat, "%d %s %s %s %s %s %s %s", 
			count, sg->nick, 
			buff2, _GMKs(sg->filesize),
			sg->starttime ? "D" : "W",
			buff, buff1,
			base_name(sg->filename)));
	}
	for (sg = napster_sendqueue; sg; sg = sg->next, count++)
	{
		char buff[80];
		char buff1[80];
		char buff2[80];
		double perc = 0.0;
		if (count == 1)
		{
			nap_put("%s", cparse("%GUploads", NULL));
			nap_put("%s", cparse("%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K%n%W%n%K", NULL, NULL));
		}
		if (sg->starttime)
			sprintf(buff, "%2.3f", sg->received / 1024.0 / (snow - sg->starttime));
		else
			strcpy(buff, "N/A");
		perc = (100.0 * (((double)(sg->received + sg->resume)) / (double)sg->filesize));
		sprintf(buff1, "%4.1f%%", perc);
		sprintf(buff2, "%4.2f", _GMKv(sg->filesize));
		nap_put("%s", cparse(dformat, "%d %s %s %s %s %s %s %s", 
			count, sg->nick, 
			buff2, _GMKs(sg->filesize),
			sg->starttime ? "U" : "W",
			buff, buff1,
			base_name(sg->filename)));
	}
}

void nap_request (IrcCommandDll *intp, char *command, char *args, char *subargs, char *helparg)
{
char *nick, *filen, *comm;

	if (!my_stricmp(command, "nrequest"))
	{
		nick = next_arg(args, &args);
		filen = new_next_arg(args, &args);
		if (nick && filen && *filen)
		{
			GetFile *new;
			send_ncommand(CMDS_REQUESTFILE, "%s \"%s\"", nick, filen);
			new = new_malloc(sizeof(GetFile));
			new->nick = m_strdup(nick);
			new->filename = m_strdup(filen);
			new->next = getfile_struct;
			getfile_struct = new;
		}
	}
	else if (!my_stricmp(command, "nget"))
	{
		int	i = 0, count = 1;
		FileStruct *sf = NULL;
		while (args && *args)
		{
			int req, browse;

			sf = NULL;
			req = browse = 0;
			count = 1;
			comm = next_arg(args, &args);
			if (!my_strnicmp(comm, "-request", 3))
			{
				req = 1;
				comm = next_arg(args, &args);
			}
			else if (!my_strnicmp(comm, "-browse", 3))
			{
				browse = 1;
				comm = next_arg(args, &args);
			}

			if (comm && *comm)
				i = strtoul(comm, NULL, 10);
			if (!req && !browse)
			{
				if (file_search)
					sf = file_search;
				else
					sf = file_browse;
			} 
			else if (req)
				sf = file_search;
			else
				sf = file_browse;

			if (sf && i)
			{
				for (; sf; sf = sf->next, count++)
				{
					if (i == count)
					{
						GetFile *new;
						send_ncommand(CMDS_REQUESTFILE, "%s \"%s\"", sf->nick, sf->name);
						new = new_malloc(sizeof(GetFile));
						new->nick = m_strdup(sf->nick);
						new->filename = m_strdup(sf->name);
						new->filesize = sf->filesize;
						new->checksum = m_strdup(sf->checksum);
						new->next = getfile_struct;
						getfile_struct = new;
						return;
					}
				}
			} 
			else if (sf)
			{
				for (; sf; sf = sf->next)
					print_file(sf, count++);
				return;
			}
		}
		if (file_search)
		{
			for (sf = file_search; sf; sf = sf->next)
				print_file(sf, count++);
		} else if (file_browse)
			for (sf = file_browse; sf; sf = sf->next)
				print_file(sf, count++);
	}
}

GetFile *find_in_getfile(GetFile **fn, int remove, char *nick, char *check, char *file, int speed, int up)
{
GetFile *last, *tmp;
	last = NULL;
	for (tmp = *fn; tmp; tmp = tmp->next)
	{
		if (!my_stricmp(tmp->nick, nick))
		{
			if (check && my_stricmp(tmp->checksum, check))
				continue;
			if (file && my_stricmp(tmp->filename, file))
				continue;
			if ((speed != -1) && (tmp->speed != speed))
				continue;
			if ((up != -1) && (tmp->up != up))
				continue;
			if (remove)
			{
				if (last)
					last->next = tmp->next;
				else
					*fn = tmp->next;
			}
			return tmp;
		}
		last = tmp;
	}
	return NULL;
}

void getfile_cleanup(int snum)
{
SocketList *s;
	if ((s = get_socket(snum)) && s->info)
	{
		GetFile *f = (GetFile *)s->info;
		if ((f = find_in_getfile(&getfile_struct, 1, f->nick, f->checksum, f->filename, -1, NAP_DOWNLOAD)))
		{
			new_free(&f->nick);
			new_free(&f->filename);
			new_free(&f->realfile);
			new_free(&f->ip);
			new_free(&f->checksum);
			if (f->write > 0)
				close(f->write);
			new_free(&f);
		}
	}
	close_socketread(snum);
	build_napster_status(NULL);
	return;
}

void nap_getfile(int snum)
{
char indata[2*NAP_BUFFER_SIZE+1];
SocketList *s;
int rc;
int count = sizeof(indata) - 1;
GetFile *gf;
unsigned long nbytes = 0;

	s = get_socket(snum);
	gf = (GetFile *)get_socketinfo(snum);
	if (gf && gf->count)
	{
		char c;
		int flags = O_WRONLY;
		set_blocking(snum);
		memset(&indata, 0, 200);
		if ((rc = read(snum, &c, 1)) != 1)
			return;
		read(snum, &indata, gf->count);
		if (!isdigit(c) || !isdigit(*indata))
		{
			gf = find_in_getfile(&getfile_struct, 1, gf->nick, gf->checksum, gf->filename, -1, NAP_DOWNLOAD);
			nap_finished_file(snum, gf);
			return;
		}
		gf->count = 0;
		set_non_blocking(snum);
		gf->starttime = time(NULL);
		if (!gf->resume)
			flags |= O_CREAT;
		if (!gf->realfile || ((gf->write = open(gf->realfile, flags, 0666)) == -1))
		{
			nap_say("Error opening output file %s: %s\n", base_name(gf->realfile), strerror(errno));
			gf = find_in_getfile(&getfile_struct, 1, gf->nick, gf->checksum, gf->filename, -1, NAP_DOWNLOAD);
			nap_finished_file(snum, gf);
			return;
		}
		if (gf->resume)
			lseek(gf->write, gf->resume, SEEK_SET);	
		nap_say("%s", cparse("$0ing from $1 [$2-]", "%s %s %s", gf->resume?"Resum":"Gett", gf->nick, base_name(gf->filename)));
		add_sockettimeout(snum, 0, NULL);
		build_napster_status(NULL);
		return;
	} 
	else if (!gf)
	{
		close_socketread(snum);
		return;
	}
        if ((rc = ioctl(snum, FIONREAD, &nbytes) != -1))
	{
		if (nbytes)
		{
			count = (nbytes > count) ? count : nbytes;
			rc = read(snum, indata, count);
		} else
			rc = 0;
	}
	switch (rc)
	{
		case -1:
			nap_say("ERROR reading file [%s]", strerror(errno));
		case 0:
		{
			if (gf)
			{
				statistics.files_received++;
				statistics.filesize_received += gf->received;
				if ((gf = find_in_getfile(&getfile_struct, 1, gf->nick, gf->checksum, gf->filename, -1, NAP_DOWNLOAD)))
				{
					char speed[80];
					sprintf(speed, "%4.2fK/s", gf->received / 1024.0 / (now - gf->starttime));
					if (rc != -1)
						nap_say("%s", cparse("Finished $2ing [$3-] from $0 $1", "%s %s %s %s", gf->nick, speed, gf->resume ? "Resum":"Gett", base_name(gf->filename)));
				}
			}
			nap_finished_file(snum, gf);
			build_napster_status(NULL);
			return;
		}
		default:
			break;
	}
	write(gf->write, indata, rc);
	gf->received += rc;
	if ((gf->received+gf->resume) >= gf->filesize)
	{
		if ((gf = find_in_getfile(&getfile_struct, 1, 
			gf->nick, gf->checksum, gf->filename, -1, NAP_DOWNLOAD)))
		{
			char speed1[80];
			double speed;
			speed = gf->received / 1024.0 / (now - gf->starttime);
			sprintf(speed1, "%4.2fK/s", speed);
			if (speed > statistics.max_downloadspeed)
				statistics.max_downloadspeed = speed;
			nap_say("%s", cparse("Finished $2ing [$3-] from $0 at $1", "%s %s %s %s", gf->nick, speed1, gf->resume?"Resum":"Gett", base_name(gf->filename)));
			statistics.files_received++;
			statistics.filesize_received += gf->received;
		}
		nap_finished_file(snum, gf);
		build_napster_status(NULL);
	}
}

NAP_COMM(cmd_getfileinfo)
{
char *nick;
GetFile *gf;
char indata[2*NAP_BUFFER_SIZE+1];
int count;
	nick = next_arg(args, &args);
	count = my_atol(args);
	if ((gf = find_in_getfile(&getfile_struct, 0, nick, NULL, NULL, count, NAP_DOWNLOAD)))
	{
		sprintf(indata, "%lu", gf->filesize);
		gf->count = strlen(indata);
		write(gf->socket, "GET", 3);
	        snprintf(indata, sizeof(indata), "%s \"%s\" %lu", get_dllstring_var("napster_user"), gf->filename, gf->resume);
		write(gf->socket, indata, strlen(indata));
		add_socketread(gf->socket, gf->port, gf->write, gf->nick, nap_getfile, NULL);
		set_socketinfo(gf->socket, gf);
		add_sockettimeout(gf->socket, 180, getfile_cleanup);
	}
	return 0;	
}

NAP_COMM(cmd_getfile)
{
/*
gato242 3068149784 6699 "d:\mp3\Hackers_-_07_-_Orbital_-_Halcyon_&_On_&_On.mp3"
8b451240c17fec98ea4f63e26bd42c60 7
*/
unsigned short port;
int getfd;
int speed;
char *nick, *file, *checksum, *ip, *dir = NULL;
char *realfile = NULL;
char indata[2*NAP_BUFFER_SIZE+1];
struct sockaddr_in socka;
GetFile *gf = NULL;
struct stat st;
                
	nick = next_arg(args, &args);
	ip = next_arg(args, &args);
	port = my_atol(next_arg(args, &args));
	file = new_next_arg(args, &args);
	checksum = next_arg(args, &args);
	speed = my_atol(args);

	if (!(gf = find_in_getfile(&getfile_struct, 1, nick, checksum, file, -1, NAP_DOWNLOAD)))
	{
		nap_say("%s", "request not in getfile");
		return 0;
	}
	if (port == 0)
		port = 6699;

	getfd = socket(AF_INET, SOCK_STREAM, 0);
	socka.sin_addr.s_addr = strtoul(ip, NULL, 10);
	socka.sin_family = AF_INET;
	socka.sin_port = htons(port);

	alarm(get_int_var(CONNECT_TIMEOUT_VAR));
	if (connect(getfd, (struct sockaddr *)&socka, sizeof(struct sockaddr)) != 0) 
	{
		alarm(0);
		nap_say("ERROR connecting [%s]", strerror(errno));
		new_free(&gf->nick);
		new_free(&gf->filename);
		new_free(&gf);
		return 0;
	}
	alarm(0);
	send_ncommand(CMDS_REQUESTINFO, nick);

	gf->socket = getfd;
	gf->port = port;
	gf->ip = m_strdup(ip);	
	gf->checksum = m_strdup(checksum);
	gf->speed = atol(args);

	if (!(dir = get_dllstring_var("napster_download_dir")))
		if (!(dir = get_string_var(DCC_DLDIR_VAR)))
			dir = "~";
	snprintf(indata, sizeof(indata), "%s/%s", dir, base_name(file));

	realfile = expand_twiddle(indata);

	gf->realfile = realfile;
	if (!(stat(realfile, &st)))
		gf->resume = st.st_size;

	gf->write = -1;
	gf->next = getfile_struct;
	gf->up = NAP_DOWNLOAD;
	getfile_struct = gf;
	return 0;
}
