
#include "irc.h"
#include "struct.h"
#include "dcc.h"
#include "ircaux.h"
#include "ctcp.h"
#include "cdcc.h"
#include "input.h"
#include "status.h"
#include "lastlog.h"
#include "screen.h"
#include "vars.h" 
#include "misc.h"
#include "output.h"
#include "module.h"
#include "hook.h"
#include "hash2.h"
#include "napster.h"
#include "md5.h"

#include "bsdglob.h"
#include "modval.h"

#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define DEFAULT_FILEMASK "*.mp3"

typedef struct _AUDIO_HEADER {
	int IDex;
	int ID;
	int layer;
	int protection_bit;
	int bitrate_index;
	int sampling_frequency;
	int padding_bit;
	int private_bit;
	int mode; /* 0 = STEREO 1 = Joint 2 = DUAL 3 = Mono */
	int mode_extension;
	int copyright;
	int original;
	int emphasis;
	int stereo;
	int jsbound;
	int sblimit;
	int true_layer;
	int framesize;
} AUDIO_HEADER;

Stats statistics = { 0, };


static unsigned char _buffer[32];
static int _bptr = 0;

static int current_sending = 0;


typedef struct _files {
	struct _files	*next;
	char		*filename;
	char		*checksum;
	unsigned long	filesize;
	time_t		time;
	int		bitrate;
	int		freq;
	int		stereo;
} Files;

Files *fserv_files = NULL;
GetFile *napster_sendqueue = NULL;


static char *convertnap_dos(char *str)
{
register char *p;
	for (p = str; *p; p++)
		if (*p == '/')
			*p = '\\';
	return str;
}

static char *convertnap_unix(char *arg)
{
register char *x = arg;
	while (*x)
	{
		if (*x == '\\')
			*x = '/';
		x++;
	}
	return arg;
}


char *mode_str(int mode)
{
	switch(mode)
	{	
		case 0:
			return "Stereo";
		case 1:
			return "Joint-Stereo";
		case 2:
			return "Dual-Channel";
		case 3:
			return "Mono";
	}
	return empty_string;
}

char *print_time(time_t input)
{
	static char	buff[40];
	time_t		seconds,
			minutes;
	seconds = input;
	minutes = seconds / 60;
	seconds = seconds % 60;
	sprintf(buff, "%02u:%02u", (unsigned int)minutes, (unsigned int)seconds);
	return buff;
}                                        

char *make_mp3_string(FILE *fp, Files *f, char *fs, char *dirbuff)
{
	static char	buffer[2*NAP_BUFFER_SIZE+1];
	char		*s,
			*loc,
			*p,
			*fn;

	if (!fs || !*fs)
		return empty_string;
	memset(buffer, 0, sizeof(buffer));

	loc = LOCAL_COPY(f->filename);
	fn = strrchr(loc, '/');
	*fn++ = 0;
	if ((p = strrchr(loc, '/')))
		*p++ = 0;
	/* fn should point to the filename and p to the dir */
	/* 
	 * init the dir keeper 
	 * or cmp the old dir with the new
	 */
	if (dirbuff && (!*dirbuff || strcmp(dirbuff, p)))
	{
		strcpy(dirbuff, p);
		if (fp)
			fprintf(fp, "\nDirectory [ %s ]\n", dirbuff);
		else
			return NULL;
	}
	/* size bitrate [time] filename */
	s = buffer;
	while (*fs)
	{
		if (*fs == '%')
		{
			int prec = 0, fl = 0;
			fs++;
			if (isdigit(*fs))
			{
				prec = strtol(fs, &fs, 0);
				if (*fs == '.')
					fl = strtoul(fs+1, &fs, 0);
			}
			switch(*fs)
			{
				case '%':
					*s++ = *fs;
					break;
				case 'b':
					sprintf(s, "%*u", prec, f->bitrate);
					break;
				case 's':
					if (!prec) prec = 3;
					sprintf(s, "%*.*f%s", prec, fl, _GMKv(f->filesize), _GMKs(f->filesize));
					break;
				case 't':
					strcpy(s, print_time(f->time));
					break;
				case 'T':
					strcpy(s, ltoa(f->time));
					break;
				case 'f':
					strcpy(s, fn);
					break;
				case 'F':
					strcpy(s, f->filename);
					break;
				case 'M':
					strcpy(s, f->checksum);
					break;
				case 'S':
					strcpy(s, mode_str(f->stereo));
					break;
				case 'H':
					sprintf(s, "%*.*f", prec, fl, ((double)f->freq) / ((double)1000.0));
					break;
				case 'h':
					sprintf(s, "%*u", prec, f->freq);
					break;
				default:
					*s++ = *fs;
					break;
			}
		}
		else if (*fs == '\\')
		{
			fs++;
			switch(*fs)
			{
				case 'n':
					strcpy(s, "\n");
					break;
				case 't':
					strcpy(s, "\t");
					break;
				default:
					*s++ = *fs++;
			}
		}
		else
			*s++ = *fs;
		while (*s) s++;
		fs++;
	}
	if (fp && *buffer)
		fprintf(fp, buffer);
	return buffer;
}


int read_glob_dir(char *path, int globflags, glob_t *globpat, int recurse)
{
	char	buffer[NAP_BUFFER_SIZE+1];
	
	sprintf(buffer, "%s/*", path);
	bsd_glob(buffer, globflags, NULL, globpat);
	if (recurse)
	{
		int i = 0;
		int old_glpathc = globpat->gl_pathc;
		for (i = 0; i < old_glpathc; i++)
		{
			char *fn;
			fn = globpat->gl_pathv[i];
			if (fn[strlen(fn)-1] != '/')
				continue;
			sprintf(buffer, "%s*", fn);
			bsd_glob(buffer, globflags|GLOB_APPEND, NULL, globpat);
		}
		while (i < globpat->gl_pathc)
		{
			for (i = old_glpathc, old_glpathc = globpat->gl_pathc; i < old_glpathc; i++)
			{
				char *fn;
				fn = globpat->gl_pathv[i];
				if (fn[strlen(fn)-1] != '/')
					continue;
				sprintf(buffer, "%s*", fn);
				bsd_glob(buffer, globflags|GLOB_APPEND, NULL, globpat);
			}
		}
	}
	return 0;
}

unsigned int print_mp3(char *pattern, char *format, int freq, int number, int bitrate, int md5)
{
unsigned int count = 0;
Files *new;
char dir[NAP_BUFFER_SIZE];
char *fs = NULL;
	*dir = 0;
	for (new = fserv_files; new; new = new->next)
	{
		if (!pattern || (pattern && wild_match(pattern, new->filename)))
		{
			char *p;
			p = base_name(new->filename);
			if ((bitrate != -1) && (new->bitrate != bitrate))
				continue;
			if ((freq != -1) && (new->freq != freq))
				continue;
			if (!format || !*format)
			{
				if (md5)
					put_it("\"%s\" %s %dk [%s]", p, new->checksum, new->bitrate, print_time(new->time));
				else
					put_it("\"%s\" %s %dk [%s]", p, mode_str(new->stereo), new->bitrate, print_time(new->time));
			}
			else
			{
				if ((fs = make_mp3_string(NULL, new, format, dir)))
					put_it("%s", fs);
				else
					put_it("%s", make_mp3_string(NULL, new, format, dir));
			}
		}
		if ((number > 0) && (count == number))
			break;
		count++;
	}
	return count;
}

void print_napster(IrcCommandDll *intp, char *command, char *args, char *subargs, char *help)
{
	int	count = 0;
	int	bitrate = -1;
	int	number = -1;
	int	freq = -1;
	int	md5 = 0;
	char 	*fs_output = NULL;
	char	*tmp_pat = NULL;
	
	if ((get_dllstring_var("napster_format")))
		fs_output = m_strdup(get_dllstring_var("napster_format"));
	if (args && *args)
	{
		char *tmp;
		while ((tmp = next_arg(args, &args)) && *tmp)
		{
			int len;
			len = strlen(tmp);
			if (!my_strnicmp(tmp, "-BITRATE", len))
			{
				if ((tmp = next_arg(args, &args)))
					bitrate = (unsigned int) my_atol(tmp);
			}
			else if (!my_strnicmp(tmp, "-COUNT", len))
			{
				if ((tmp = next_arg(args, &args)))
					number = (unsigned int) my_atol(tmp);
			} 
			else if (!my_strnicmp(tmp, "-FREQ", 3))
			{
				if ((tmp = next_arg(args, &args)))
					freq = (unsigned int)my_atol(tmp);
			} 
			else if (!my_strnicmp(tmp, "-MD5", 3))
			{
				md5 = 1;
			} 
			else if (!my_strnicmp(tmp, "-FORMAT", 3))
			{
				if ((tmp = new_next_arg(args, &args)))
					malloc_strcpy(&fs_output, tmp);
			} 
			else
			{
				count += print_mp3(tmp, fs_output, freq, number, bitrate, md5);
				m_s3cat(&tmp_pat, " ", tmp);
			}
		}
	}
	else
		count += print_mp3(NULL, fs_output, freq, number, bitrate, md5);
	
	nap_say("Found %d files matching \"%s\"", count, tmp_pat ? tmp_pat : "*");
	new_free(&tmp_pat);
	new_free(&fs_output);
}

int _get_input(int file, unsigned char *bp, int size)
{
	if (read(file,  bp, size) != size)
		return -1;
	return 0;
}
                                        
static inline int readsync(int file)
{
	_bptr=0;
	_buffer[0]=_buffer[1];
	_buffer[1]=_buffer[2];
	_buffer[2]=_buffer[3];
	return _get_input(file, &_buffer[3], 1);
}


static inline int _fillbfr(int file, unsigned int size)
{
	_bptr=0;
        return _get_input(file, _buffer, size);
}


static inline unsigned int _getbits(int n)
{
	unsigned int	pos,
			ret_value;

        pos = _bptr >> 3;
	ret_value = _buffer[pos] << 24 |
		    _buffer[pos+1] << 16 |
		    _buffer[pos+2] << 8 |
		    _buffer[pos+3];
        ret_value <<= _bptr & 7;
        ret_value >>= 32 - n;
        _bptr += n;
        return ret_value;
}       


/*
 * header and side info parsing stuff ******************************************
 */
static inline void parse_header(AUDIO_HEADER *header) 
{
        header->IDex=_getbits(1);
        header->ID=_getbits(1);
        header->layer=_getbits(2);
        header->protection_bit=_getbits(1);
        header->bitrate_index=_getbits(4);
        header->sampling_frequency=_getbits(2);
        header->padding_bit=_getbits(1);
        header->private_bit=_getbits(1);
        header->mode=_getbits(2);
        header->mode_extension=_getbits(2);
        if (!header->mode) 
        	header->mode_extension=0;
        header->copyright=_getbits(1);
        header->original=_getbits(1);
        header->emphasis=_getbits(2);

	header->stereo = (header->mode == 3) ? 1 : 2;
	header->true_layer = 4 - header->layer;
}

int gethdr(int file, AUDIO_HEADER *header)
{
	int	retval;

	if ((retval=_fillbfr(file, 4))) 
		return retval;

	while (_getbits(11) != 0x7ff) 
	{
		if ((retval=readsync(file))!=0) 
			return retval;
	}
	parse_header(header);
	return 0;
}

long get_bitrate(char *filename, time_t *mp3_time, unsigned int *freq_rate, unsigned long *filesize, int *stereo)
{
	short t_bitrate[2][3][15] = {{
	{0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
	{0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
	{0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}
	},{
	{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
	{0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
	{0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}
	}};

	int t_sampling_frequency[2][2][3] = {
		{ /* MPEG 2.5 samplerates */ 
  			{ 11025, 12000, 8000},
			{ 0,0,0 }
		},{ /* MPEG 2.0/1.0 samplerates */
			{ 22050 , 24000 , 16000},
			{ 44100 , 48000 , 32000}
		}
	};

	AUDIO_HEADER header;
	unsigned long btr = 0;
	int l = -1;
	struct stat	st;
	unsigned long	framesize = 0,
			totalframes = 0;
	
	if (freq_rate)
		*freq_rate =0;
	if ((l = open(filename, O_RDONLY)) == -1) 
		return 0;
	gethdr(l, &header);
	if (header.ID > 1 || header.layer > 2 || header.bitrate_index > 14) 
	{
		close(l);
		return 0;
	}
	btr = t_bitrate[header.ID][3-header.layer][header.bitrate_index];

	fstat(l, &st);
	if (t_sampling_frequency[header.IDex][header.ID][header.sampling_frequency] > 0)
		framesize = (header.ID ? 144000 : 72000) * btr / t_sampling_frequency[header.IDex][header.ID][header.sampling_frequency];
	totalframes = (st.st_size / (framesize + 1)) - 1;
	if (t_sampling_frequency[header.IDex][header.ID][header.sampling_frequency] > 0)
		*mp3_time = (time_t) (totalframes * (header.ID==0?576:1152)/t_sampling_frequency[header.IDex][header.ID][header.sampling_frequency]);
	*filesize = st.st_size;

	if (freq_rate)
		*freq_rate = t_sampling_frequency[header.IDex][header.ID][header.sampling_frequency];

	*stereo = header.mode;
	close(l);	        
	return btr;
}

off_t file_size (char *filename)
{
	struct stat statbuf;

	if (!stat(filename, &statbuf))
		return (off_t)(statbuf.st_size);
	else
		return -1;
}

unsigned int scan_mp3_dir(char *path, int recurse, int reload, int share)
{
	int	mt = 0;
	glob_t	globpat;
	int	i = 0;
	Files	*new;
	int	count = 0;
	int	r;
	char	buffer[2*NAP_BUFFER_SIZE+1];
	
	memset(&globpat, 0, sizeof(glob_t));
	read_glob_dir(path, GLOB_MARK|GLOB_NOSORT, &globpat, recurse);
	for (i = 0; i < globpat.gl_pathc; i++)
	{
		char	*fn;
		fn = globpat.gl_pathv[i];
		if (fn[strlen(fn)-1] == '/')
			continue;
		if (!(mt = wild_match(DEFAULT_FILEMASK, fn)))
			continue;
		if (reload && find_in_list((List **)&fserv_files, globpat.gl_pathv[i], 0))
			continue;
		new = (Files *) new_malloc(sizeof(Files));
		new->filename = m_strdup(fn);
		new->bitrate = get_bitrate(fn, &new->time, &new->freq, &new->filesize, &new->stereo);
		if (new->filesize && ((r = open(fn, O_RDONLY)) != -1))
		{
			int di = 0;
#if !defined(WINNT)
			char *m;
#endif
			unsigned long size = 300000;
			unsigned char digest[16];			
			md5_state_t state;
			int rc;
					
			md5_init(&state);
			if (new->filesize < size)
				size = new->filesize;
#if defined(WINNT)
			while (size)
			{
				unsigned char md5_buff[4 * NAP_BUFFER_SIZE+1];
				rc = (size >= (4 * NAP_BUFFER_SIZE)) ?  4 * NAP_BUFFER_SIZE : size;
				rc = read(r, md5_buff, rc);
				md5_append(&state, (const md5_byte_t *) (unsigned char *)md5_buff, rc);
				if (size >= 4 * NAP_BUFFER_SIZE)
					size -= rc;
				else
					size = 0;
			}						
			md5_finish(&state, digest);
			{
#else
			if ((m = mmap((void *)0, size, PROT_READ, MAP_PRIVATE, r, 0)) != MAP_FAILED)
			{
				md5_append(&state, (const md5_byte_t *) (unsigned char *)m, size);
				md5_finish(&state, digest);
				munmap(m, size);
#endif
				memset(buffer, 0, 200);
				for (di = 0, rc = 0; di < 16; ++di, rc += 2)
					sprintf(&buffer[rc], "%02x", digest[di]);
				new->checksum = m_strdup(buffer);
				add_to_list((List **)&fserv_files, (List *)new);
				statistics.total_files++;
				statistics.total_filesize += new->filesize;
				count++;
				if (!(count % 25))
				{
					lock_stack_frame();
					io("scan_mp3_dir");
					unlock_stack_frame();
					build_napster_status(NULL);
				}
				if (share && (nap_socket != -1))
				{
					sprintf(buffer, "\"%s\" %s %lu %u %u %lu", new->filename, 
						new->checksum, new->filesize, new->bitrate, new->freq, new->time);
					send_ncommand(CMDS_ADDFILE, convertnap_dos(buffer));
					statistics.shared_files++;
					statistics.shared_filesize += new->filesize;
				}
			} 
			close(r);
		}
		else
		{
			new_free(&new->filename);
			new_free(&new);
		}
	}
	bsd_globfree(&globpat);
	return count;
}

void load_napserv(IrcCommandDll *intp, char *command, char *args, char *subargs, char *help)
{
	char	*path = NULL;
	int	recurse = 1;
	char	*pch;
	int	count = 0;
	int	reload = 0;
	int	share = 0;
		
	if (command && !my_stricmp(command, "NRELOAD"))
		reload = 1;
	if (args && *args)
	{
		while ((path = next_arg(args, &args)) && *path)
		{
			int len = strlen(path);
			if (!my_strnicmp(path, "-recurse", len))
			{
				recurse ^= 1;
				continue;
			}
			if (!my_strnicmp(path, "-share", len))
			{
				share ^= 1;
				continue;
			}
			count += scan_mp3_dir(path, recurse, reload, share);
		}
		goto fs_print_out;
	}

	path = get_dllstring_var("napster_dir");

	if (!path || !*path)
	{
		nap_say("No path. /set napster_dir first.");
		return;
	}

	pch = LOCAL_COPY(path);
	while ((path = next_arg(pch, &pch)))
		count += scan_mp3_dir(path, recurse, reload, share);
fs_print_out:
	build_napster_status(NULL);
	if (!fserv_files || !count)
		nap_say("Could not read dir");
	else
		nap_say("Found %d files%s", count, share ? "" : ". To share these type /nshare");
	return;
}

void share_napster(IrcCommandDll *intp, char *command, char *args, char *subargs, char *help)
{
unsigned long count = 0;
char buffer[2*NAP_BUFFER_SIZE+1];
Files	*new;
	for (new = fserv_files; new && (nap_socket != -1); new = new->next, count++)
	{
		sprintf(buffer, "\"%s\" %s %lu %u %u %lu", new->filename, 
			new->checksum, new->filesize, new->bitrate, new->freq, new->time);
		send_ncommand(CMDS_ADDFILE, convertnap_dos(buffer));
		if (!(count % 20))
		{
			lock_stack_frame();
			io("share napster");
			unlock_stack_frame();
			build_napster_status(NULL);
		}
		statistics.shared_files++;
		statistics.shared_filesize += new->filesize;
	}
	build_napster_status(NULL);
	nap_say("%s", cparse("Sharing $0 files", "%l", count));
}

int nap_finished_file(int snum, GetFile *gf)
{
SocketList *s = NULL;
	if (snum != -1)
		s = get_socket(snum);
	if (gf && ((GetFile *)s->info == gf))
	{
		s->is_write = 0;
		if (gf->write > 0)
			close(gf->write);
		new_free(&gf->nick);
		new_free(&gf->filename);
		new_free(&gf->checksum);
		new_free(&gf->realfile);
		new_free(&gf->ip);
		if (gf->up == NAP_UPLOAD)
			current_sending--;
		new_free(&gf);
	}
	if (snum != -1)
		close_socketread(snum);
	return 0;
}

void sendfile_timeout(int snum)
{
SocketList *s;
GetFile *f = NULL;
	if ((s = get_socket(snum)) && s->info)
	{
		f = (GetFile *)s->info;
		f = find_in_getfile(&napster_sendqueue, 1, f->nick, NULL, f->filename, -1, NAP_UPLOAD);
	}
	if (f)
		nap_say("%s", cparse("Send to $0 timed out [$1-", "%s %s", f->nick, strerror(errno)));
	nap_finished_file(snum, f);
	build_napster_status(NULL);
	return;
}

int clean_queue(GetFile **gf, int timeout)
{
GetFile *ptr;
int count = 0;
	if (!gf || !*gf || timeout <= 0)
		return 0;
	ptr = *gf;
	while (ptr)
	{

		if (ptr->addtime && ((ptr->addtime + timeout) >= now))
		{
			if (!(ptr = find_in_getfile(gf, 1, ptr->nick, NULL, ptr->filename, -1, NAP_UPLOAD)))
				continue;
			new_free(&ptr->nick);
			new_free(&ptr->filename);
			new_free(&ptr->checksum);
			new_free(&ptr->realfile);
			new_free(&ptr->ip);
			if (ptr->up == NAP_UPLOAD)
				current_sending--;
			new_free(&ptr);
			ptr = *gf;
			count++;
		} 
		else 
			ptr = ptr->next;
	}
	return count;
}

NAP_COMM(cmd_filerequest)
{
char *nick;
Files *new = NULL;
char *filename;
int count = 0;

	nick = next_arg(args, &args);
	filename = new_next_arg(args, &args);
	if (!filename || !*filename)
		return 0;
	convertnap_unix(filename);
	for (new = fserv_files; new; new = new->next)
	{
		if (!strcmp(filename, new->filename))
			break;
	}
	if (new)
	{
		char buffer[2*NAP_BUFFER_SIZE+1];
		GetFile *gf;
		for (gf = napster_sendqueue; gf; gf = gf->next)
		{
			if (!gf->filename)
			{
				nap_say("ERROR in cmd_filerequest. gf->filename is null");
				return 0;
				
			}
			count++;
			if (!strcmp(filename, gf->filename) && !strcmp(nick, gf->nick))
			{
				nap_say("%s", cparse("$0 is already queued for $1-", "%s %s", gf->nick, gf->filename));
				return 0;
			}
		}
		if (count > get_dllint_var("napster_send_limit"))
		{
#if 0
			sprintf(buffer, "%s \"%s\"", nick, new->filename);
			send_ncommand(609, convertnap_dos(buffer));
#endif
			return 0;
		}
		nap_say("%s", cparse("$0 has requested [$1-]", "%s %s", nick, base_name(filename)));
		sprintf(buffer, "%s \"%s\"", nick, new->filename);
		send_ncommand(CMDS_REQUESTINFO, nick);
		send_ncommand(CMDS_FILEINFO, convertnap_dos(buffer));
		gf = new_malloc(sizeof(GetFile));
		gf->nick = m_strdup(nick);
		gf->filename = m_strdup(new->filename);
		gf->write = open(new->filename, O_RDONLY);
		gf->filesize = new->filesize;
		gf->next = napster_sendqueue;
		gf->up = NAP_UPLOAD;
		gf->addtime = time(NULL);
		napster_sendqueue = gf;
		current_sending++;
		clean_queue(&napster_sendqueue, 300);
	} 
	return 0;
}

void napfile_read(int snum)
{
GetFile *gf;
unsigned char buffer[NAP_BUFFER_SIZE+1];
int rc, numread;
SocketList *s;
	s = get_socket(snum);
	if (!(gf = (GetFile *)get_socketinfo(snum)))
	{
		unsigned char buff[2*NAP_BUFFER_SIZE+1];
		unsigned char fbuff[2*NAP_BUFFER_SIZE+1];
		char *nick, *filename, *args;

		recv(snum, buff, 4, MSG_PEEK);
		if (!*buff || strncmp(buff, "GET", 3))
		{
			close_socketread(snum);
			return;
		}
		read(snum, buff, 3);
		rc = read(snum, buff, sizeof(buff)-1);
		buff[rc] = 0;
		args = &buff[0];
		nick = next_arg(args, &args);
		filename = new_next_arg(args, &args);
		strcpy(fbuff, filename);
		convertnap_unix(fbuff);
		if (!(gf = find_in_getfile(&napster_sendqueue, 0, nick, NULL, fbuff, -1, NAP_UPLOAD)) || (gf->write == -1))
		{
			memset(buff, 0, 80);
			if (!gf)
				sprintf(buff, "0INVALID REQUEST");
			else
				sprintf(buff, "0FILE NOT FOUND");
			write(snum, buff, strlen(buffer));
			nap_finished_file(snum, gf);
			return;
		}
		gf->resume = strtoul(args, NULL, 0);
		if (gf->resume >= gf->filesize)
		{
			gf = find_in_getfile(&napster_sendqueue, 1, nick, NULL, fbuff, -1, NAP_UPLOAD);
			nap_finished_file(snum, gf);
			return;
		}
		lseek(gf->write, SEEK_SET, gf->resume);
		set_socketinfo(snum, gf);
		memset(buff, 0, 80);
		sprintf(buff, "%lu", gf->filesize);
		write(snum, buff, strlen(buff));
		s->func_write = s->func_read;
		s->is_write = s->is_read;
		nap_say("%s", cparse("$0ing file to $1 [$2-]", "%s %s %s", gf->resume ? "Resum" : "Send", gf->nick, base_name(gf->filename)));
		add_sockettimeout(snum, 0, NULL);
		set_non_blocking(snum);
		build_napster_status(NULL);
		return;
	} else if (!gf->starttime)
		gf->starttime = now;

	gf->addtime = now;
	numread = read(gf->write, buffer, sizeof(buffer)-1);
	switch(numread)
	{
		case -1:
		case 0:
		{
			if ((gf = find_in_getfile(&napster_sendqueue, 1, gf->nick, NULL, gf->filename, -1, NAP_UPLOAD)))
			{
				if ((gf->received + gf->resume) >= gf->filesize)
				{
					double speed;
					char speed1[80];
					statistics.files_served++;
					statistics.filesize_served += gf->received;
					speed = gf->received / 1024.0 / (time(NULL) - gf->starttime);
					if (speed > statistics.max_uploadspeed)
						statistics.max_uploadspeed = speed;
					sprintf(speed1, "%4.2fK/s", speed);
					nap_say("%s", cparse("Finished Sending $0 [$1-]", "%s %s %s", gf->nick, speed1, base_name(gf->filename)));
				}
				else
					nap_say("%s", cparse("Error receiving $2 from $0 [$3-]", "%s %s \"%s\"", gf->nick, base_name(gf->filename), gf->nick));
			}
			nap_finished_file(snum, gf);
			build_napster_status(NULL);
			break;
		}
		default:
		{
			rc = send(snum, buffer, numread, 0);	
			if (rc != numread)
			{
				if (rc == -1)
				{
					if (errno == EWOULDBLOCK || errno == ENOBUFS)
						lseek(gf->write, -numread, SEEK_CUR);
					else
					{
						if ((gf = find_in_getfile(&napster_sendqueue, 1, gf->nick, NULL, gf->filename, -1, NAP_UPLOAD)))
							nap_say("%s", cparse("Closed File send to $0 [$1-]", "%s %s", gf->nick, strerror(errno)));
						nap_finished_file(snum, gf);
						build_napster_status(NULL);
					}
					return;
				}
				else 
					lseek(gf->write, -(numread - rc), SEEK_CUR);
			}
			gf->received += rc;
			if (!(gf->received % (10 * (sizeof(buffer)-1))))
				build_napster_status(NULL);
		}
	}
}

void naplink_handlelink(int snum)
{
struct  sockaddr_in     remaddr;
int sra = sizeof(struct sockaddr_in);
int sock = -1;

	if ((sock = accept(snum, (struct sockaddr *) &remaddr, &sra)) > -1)
	{
		set_napster_socket(sock);
		add_socketread(sock, snum, 0, inet_ntoa(remaddr.sin_addr), napfile_read, NULL);
		add_sockettimeout(sock, 180, sendfile_timeout);
		write(sock, "\n", 1);
	}
}
